"""
pygame-menu
https://github.com/ppizarror/pygame-menu

DECORATOR
Generic decorator, adds additional images, polygons or text to the object.
"""
# File constants no. 2000

__all__ = ['Decorator']

from math import pi
import math

import pygame
import pygame_menu
import pygame_menu.menu
import pygame.draw as pydraw
import pygame.gfxdraw as gfxdraw

from pygame_menu._base import Base
from pygame_menu.font import FontType
from pygame_menu.utils import assert_list_vector, assert_color, make_surface, \
    assert_vector, uuid4, warn

from pygame_menu._types import List, Tuple2NumberType, ColorInputType, Tuple, \
    Any, Dict, Union, NumberType, Tuple2IntType, Optional, Callable, NumberInstance, \
    CallableNoArgsType

# Decoration constants
DECORATION_ARC = 2000
DECORATION_BASEIMAGE = 2001
DECORATION_BEZIER = 2002
DECORATION_CALLABLE = 2003
DECORATION_CALLABLE_NO_ARGS = 2015
DECORATION_CIRCLE = 2004
DECORATION_ELLIPSE = 2005
DECORATION_FILL = 2020
DECORATION_LINE = 2006
DECORATION_NONE = 2007
DECORATION_PIE = 2008
DECORATION_PIXEL = 209
DECORATION_POLYGON = 2010
DECORATION_RECT = 2011
DECORATION_SURFACE = 2012
DECORATION_TEXT = 2013
DECORATION_TEXTURE_POLYGON = 2014

DECOR_TYPE_PREV = 'prev'
DECOR_TYPE_POST = 'post'


# noinspection PyProtectedMember
class Decorator(Base):
    """
    Decorator class.

    :param obj: Object to decorate
    :param decorator_id: ID of the decorator
    :param verbose: Enable/disable verbose mode (warnings/errors)
    """
    _coord_cache: Dict[str, Tuple[int, int, Union[Tuple[Tuple2NumberType, ...], Tuple2NumberType]]]  # centerx, centery, coords
    _cache_last_status: Dict[str, Tuple[int, int, int, int, int, int]]
    _cache_needs_update: Dict[str, bool]
    _cache_surface: Dict[str, Optional['pygame.Surface']]
    _decor: Dict[str, List[Tuple[int, str, Any]]]  # type, id, data
    _decor_enabled: Dict[str, bool]
    _decor_prev_id: List[str]
    _obj: Union['pygame_menu.widgets.Widget', 'pygame_menu._scrollarea.ScrollArea', 'pygame_menu.Menu']
    _post_enabled: bool
    _prev_enabled: bool
    cache: bool

    def __init__(
        self,
        obj: Union['pygame_menu.widgets.Widget', 'pygame_menu._scrollarea.ScrollArea', 'pygame_menu.Menu'],
        decorator_id: str = '',
        verbose: bool = True
    ) -> None:
        super(Decorator, self).__init__(object_id=decorator_id, verbose=verbose)

        self._coord_cache = {}
        self._decor = {DECOR_TYPE_PREV: [], DECOR_TYPE_POST: []}
        self._decor_prev_id = []  # Stores all decoration prev ids
        self._obj = obj
        self._decor_enabled = {}

        self._prev_enabled = True
        self._post_enabled = True

        # If True, enables surface cache. This is intended to be used if there's
        # many decorations in the object (for example, 400). This is an expensive
        # method anyway because surface is called many times. See the following
        # rendering times to guess how much does a decoration takes time to
        # render 1000 times (object: button)
        # 100 decoration, no cache:     0.214
        # 100 decoration, with cache:   0.646
        # 300 decoration, no cache:     0.581
        # 300 decoration, with cache:   0.606
        # 1000 decoration, no cache:    2.228
        # 1000 decoration, with cache:  0.615
        # 10000 decoration, no cache:   20.430
        # 10000 decoration, with cache: 0.599
        self.cache = False

        # Previous (surf.width, surf.height, rect.x, rect.y, rect.centerx, rect.centery
        self._cache_last_status = {DECOR_TYPE_PREV: (0, 0, 0, 0, 0, 0),
                                   DECOR_TYPE_POST: (0, 0, 0, 0, 0, 0)}
        self._cache_needs_update = {DECOR_TYPE_PREV: False, DECOR_TYPE_POST: False}
        self._cache_surface = {DECOR_TYPE_PREV: None, DECOR_TYPE_POST: None}

    def __copy__(self) -> 'Decorator':
        """
        Copy method.

        :return: Raises copy exception
        """
        raise _DecoratorCopyException('Decorator class cannot be copied')

    def __deepcopy__(self, memodict: Dict) -> 'Decorator':
        """
        Deep-copy method.

        :param memodict: Memo dict
        :return: Raises copy exception
        """
        raise _DecoratorCopyException('Decorator class cannot be deep-copied')

    def _add_decor(self, decortype: int, prev: bool, data: Any) -> str:
        """
        Adds a decoration.

        :param decortype: Decoration type
        :param prev: To prev or post
        :param data: Data of the decoration
        :return: ID of the decoration
        """
        decor_id = uuid4()

        if prev:
            assert self._prev_enabled, 'prev decorators are not enabled'
            self._decor[DECOR_TYPE_PREV].append((decortype, decor_id, data))
            self._decor_prev_id.append(decor_id)
        else:
            assert self._post_enabled, 'post decorators are not enabled'
            self._decor[DECOR_TYPE_POST].append((decortype, decor_id, data))

        # Force surface cache update
        if hasattr(self._obj, 'force_menu_surface_cache_update'):
            self._obj.force_menu_surface_cache_update()

        # Forces cache update
        self._cache_needs_update[DECOR_TYPE_PREV if prev else DECOR_TYPE_POST] = True

        # Check sizes
        if self._total_decor() >= 300 and not self.cache:
            if self._verbose:
                warn('cache is recommended if the total number of decorations exceeds 300')

        # Set automatically as enabled
        self._decor_enabled[decor_id] = True

        return decor_id

    def _add_none(self, prev: bool = True) -> str:
        """
        Adds a none decorator.

        :param prev: If ``True`` draw previous the object, else draws post
        :return: ID of the decoration
        """
        return self._add_decor(DECORATION_NONE, prev, None)

    def _total_decor(self) -> int:
        """
        Return total number of decorations.
        """
        return len(self._decor[DECOR_TYPE_PREV]) + len(self._decor[DECOR_TYPE_POST])

    def force_cache_update(self, prev: Optional[bool] = None) -> 'Decorator':
        """
        Forces cache update.

        :param prev: Update the previous or post surface cache. If ``None`` forces both caches to update
        :return: Self reference
        """
        if prev is None:
            self.force_cache_update(True)
            self.force_cache_update(False)
            return self
        self._cache_needs_update[DECOR_TYPE_PREV if prev else DECOR_TYPE_POST] = True
        return self

    def add_polygon(
        self,
        coords: Union[List[Tuple2NumberType], Tuple[Tuple2NumberType, ...]],
        color: ColorInputType,
        filled: bool,
        width: int = 0,
        prev: bool = True,
        gfx: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a polygon.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param coords: Coordinate list, being ``(0, 0)`` the center of the object
        :param color: Color of the polygon
        :param filled: If ``True`` fills the polygon with the given color
        :param width: Line border width. Only valid if ``filled=False``
        :param prev: If ``True`` draw previous the object, else draws post
        :param gfx: If ``True`` uses pygame gfxdraw instead of draw
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert len(coords) >= 3
        assert isinstance(filled, bool)
        assert isinstance(width, int) and width >= 0
        if filled:
            assert width == 0, 'width must be 0 if the polygon is filled'
            assert gfx, 'only gfxdraw support filled polygon, then gfx should be True'
        else:
            if width != 0 and gfx:
                gfx = False  # gfx don't support width
        return self._add_decor(
            DECORATION_POLYGON, prev, (tuple(coords), color, filled, width, gfx, kwargs)
        )

    def add_bezier(
        self,
        coords: Union[List[Tuple2NumberType], Tuple[Tuple2NumberType, ...]],
        color: ColorInputType,
        steps: int = 5,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a Bézier curve.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param coords: Coordinate list, being ``(0, 0)`` the center of the object
        :param color: Color of the polygon
        :param steps: Interpolation steps
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert len(coords) >= 3
        assert isinstance(steps, int) and steps >= 1
        return self._add_decor(
            DECORATION_BEZIER, prev, (tuple(coords), color, steps, kwargs)
        )

    def add_circle(
        self,
        x: NumberType,
        y: NumberType,
        radius: NumberType,
        color: ColorInputType,
        filled: bool,
        width: int = 0,
        prev: bool = True,
        gfx: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a circle.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param radius: Circle radius in px
        :param color: Color of the polygon
        :param filled: If ``True`` fills the polygon with the given color
        :param width: Line border width. Only valid if ``filled=False``
        :param prev: If ``True`` draw previous the object, else draws post
        :param gfx: If ``True`` uses pygame gfxdraw instead of draw
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert isinstance(radius, NumberInstance) and radius > 0
        assert isinstance(filled, bool)
        assert isinstance(width, int) and width >= 0
        if filled:
            assert width == 0, 'width must be 0 if the circle is filled'
        else:
            if width != 0 and gfx:
                gfx = False  # gfx don't support width
        return self._add_decor(
            DECORATION_CIRCLE, prev,
            (tuple(coords), int(radius), color, filled, width, gfx, kwargs)
        )

    def add_arc(
        self,
        x: NumberType,
        y: NumberType,
        radius: NumberType,
        init_angle: NumberType,
        final_angle: NumberType,
        color: ColorInputType,
        width: int = 0,
        prev: bool = True,
        gfx: bool = True,
        **kwargs
    ) -> str:
        """
        Adds an arc.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param radius: Circle radius in px
        :param init_angle: Initial angle in degrees, from ``0`` to ``360``
        :param final_angle: Final angle in degrees, from ``0`` to ``360``
        :param color: Color of the polygon
        :param width: Line border width. Only valid if ``filled=False``
        :param prev: If ``True`` draw previous the object, else draws post
        :param gfx: If ``True`` uses pygame gfxdraw instead of draw
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert isinstance(radius, NumberInstance) and radius > 0
        assert isinstance(init_angle, NumberInstance)
        assert isinstance(final_angle, NumberInstance)
        assert isinstance(width, int) and width >= 0
        assert init_angle != final_angle
        return self._add_decor(
            DECORATION_ARC, prev,
            (tuple(coords), int(radius), init_angle, final_angle, color, width,
             gfx, kwargs)
        )

    def add_pie(
        self,
        x: NumberType,
        y: NumberType,
        radius: NumberType,
        init_angle: NumberType,
        final_angle: NumberType,
        color: ColorInputType,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds an unfilled pie.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param radius: Circle radius in px
        :param init_angle: Initial angle in degrees, from ``0`` to ``360``
        :param final_angle: Final angle in degrees, from ``0`` to ``360``
        :param color: Color of the polygon
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert isinstance(radius, NumberInstance) and radius > 0
        assert isinstance(init_angle, NumberInstance)
        assert isinstance(final_angle, NumberInstance)
        assert init_angle != final_angle
        return self._add_decor(
            DECORATION_PIE, prev,
            (tuple(coords), int(radius), init_angle, final_angle, color, kwargs)
        )

    def add_surface(
        self,
        x: NumberType,
        y: NumberType,
        surface: 'pygame.Surface',
        prev: bool = True,
        centered: bool = False,
        **kwargs
    ) -> str:
        """
        Adds a surface.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param surface: Surface
        :param prev: If ``True`` draw previous the object, else draws post
        :param centered: If ``True`` the surface is centered
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        assert isinstance(surface, pygame.Surface)
        return self._add_decor(
            DECORATION_SURFACE, prev,
            (tuple(coords), surface, centered, kwargs)
        )

    def add_baseimage(
        self,
        x: NumberType,
        y: NumberType,
        image: 'pygame_menu.BaseImage',
        prev: bool = True,
        centered: bool = False,
        **kwargs
    ) -> str:
        """
        Adds a :py:class:`pygame_menu.baseimage.BaseImage` object.

        .. note::

            If your :py:class:`pygame_menu.baseimage.BaseImage` object changes over time
            set ``decorator.cache=False`` or force cache manually by calling
            :py:meth:`pygame_menu._decorator.Decorator.force_cache_update`.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param image: ``BaseImage`` object
        :param prev: If ``True`` draw previous the object, else draws post
        :param centered: If ``True`` the image is centered
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        assert isinstance(image, pygame_menu.BaseImage)
        return self._add_decor(
            DECORATION_BASEIMAGE, prev, (tuple(coords), image, centered, kwargs)
        )

    def add_rect(
        self,
        x: NumberType,
        y: NumberType,
        rect: 'pygame.Rect',
        color: ColorInputType,
        width: int = 0,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a BaseImage object.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param rect: Rect to draw
        :param color: Color of the rect
        :param width: Border width of the rect. If ``0`` draw a filled rectangle
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert isinstance(width, int) and width >= 0
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert isinstance(rect, pygame.Rect)
        return self._add_decor(
            DECORATION_RECT, prev, (tuple(coords), rect, color, width, kwargs)
        )

    def add_rectangle(
        self,
        x: NumberType,
        y: NumberType,
        width: NumberType,
        height: NumberType,
        color: ColorInputType,
        border: int = 0,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a BaseImage object.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param width: Rectangle width
        :param height: Rectangle height
        :param color: Color of the rectangle
        :param border: Border width of the rectangle. If ``0`` draw a filled rectangle
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert isinstance(width, NumberInstance) and width > 0
        assert isinstance(height, NumberInstance) and height > 0
        rect = pygame.Rect(0, 0, width, height)
        return self.add_rect(x, y, rect, color, border, prev, **kwargs)

    def add_text(
        self,
        x: NumberType,
        y: NumberType,
        text: str,
        font: FontType,
        size: int,
        color: ColorInputType,
        prev: bool = True,
        antialias=True,
        centered=False,
        **kwargs
    ) -> str:
        """
        Adds a text.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param text: Text to draw
        :param font: Font path or pygame object
        :param size: Size of the font to render
        :param color: Font color
        :param prev: If ``True`` draw previous the object, else draws post
        :param antialias: Font antialias enabled
        :param centered: If ``True`` the text is centered
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        text = str(text)
        font_obj = pygame_menu.font.get_font(font, size)
        color = assert_color(color)
        surface_font = font_obj.render(text, antialias, color)
        surface = make_surface(
            width=surface_font.get_width(),
            height=surface_font.get_height(),
            alpha=True
        )
        surface.blit(surface_font, (0, 0))
        return self._add_decor(
            DECORATION_TEXT, prev, (tuple(coords), surface, centered, kwargs)
        )

    def add_ellipse(
        self,
        x: NumberType,
        y: NumberType,
        rx: NumberType,
        ry: NumberType,
        color: ColorInputType,
        filled: bool,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds an ellipse.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param rx: Horizontal radius of the ellipse
        :param ry: Vertical radius of the ellipse
        :param color: Color of the polygon
        :param filled: If ``True`` fills the polygon with the given color
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        assert isinstance(rx, NumberInstance) and rx > 0
        assert isinstance(ry, NumberInstance) and ry > 0
        assert isinstance(filled, bool)
        return self._add_decor(
            DECORATION_ELLIPSE, prev, (tuple(coords), rx, ry, color, filled, kwargs)
        )

    def add_pixel(
        self,
        x: NumberType,
        y: NumberType,
        color: ColorInputType,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a pixel.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: X position in px, being ``0`` the center of the object
        :param y: Y position in px, being ``0`` the center of the object
        :param color: Color of the pixel
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        coords = [(x, y)]
        assert_list_vector(coords, 2)
        color = assert_color(color)
        return self._add_decor(
            DECORATION_PIXEL, prev, (tuple(coords), color, kwargs)
        )

    def add_callable(
        self,
        fun: Union[Callable[['pygame.Surface', Any], Any], CallableNoArgsType],
        prev: bool = True,
        pass_args: bool = True
    ) -> str:
        """
        Adds a callable method. The function receives the surface and the object;
        for example, if adding to a widget:

        .. code-block:: python

            fun(surface, object)

        .. note::

            If your callable function changes over time set ``decorator.cache=False``
            or force cache manually by calling Decorator method
            :py:meth:`pygame_menu._decorator.Decorator.force_cache_update`. Also,
            the object should force the menu surface cache to update.

        :param fun: Function
        :param prev: If ``True`` draw previous the object, else draws post
        :param pass_args: If ``False`` function is called without (surface, object) as args
        :return: ID of the decoration
        """
        assert callable(fun), 'fun must be a callable type'
        assert isinstance(pass_args, bool)
        if pass_args:
            return self._add_decor(DECORATION_CALLABLE, prev, fun)
        else:
            return self._add_decor(DECORATION_CALLABLE_NO_ARGS, prev, fun)

    def add_textured_polygon(
        self,
        coords: Union[List[Tuple2NumberType], Tuple[Tuple2NumberType, ...]],
        texture: Union['pygame.Surface', 'pygame_menu.BaseImage'],
        tx: int = 0,
        ty: int = 0,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a textured polygon.

        .. note::

            If your :py:class:`pygame_menu.baseimage.BaseImage` object changes over
            time set ``decorator.cache=False`` or force cache manually by calling
            :py:class:`pygame_menu._decorator.Decorator.force_cache_update`.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param coords: Coordinate list, being ``(0, 0)`` the center of the object
        :param texture: Texture (Surface) or Baseimage object
        :param tx: X offset of the texture in px
        :param ty: Y offset of the texture in px
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert_list_vector(coords, 2)
        assert len(coords) >= 3
        assert isinstance(texture, (pygame.Surface, pygame_menu.BaseImage))
        assert isinstance(tx, int) and isinstance(ty, int)
        return self._add_decor(
            DECORATION_TEXTURE_POLYGON, prev, (tuple(coords), texture, tx, ty, kwargs)
        )

    def add_line(
        self,
        pos1: Tuple2NumberType,
        pos2: Tuple2NumberType,
        color: ColorInputType,
        width: int = 1,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a line.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param pos1: Position 1 (x1, y1)
        :param pos2: Position 2 (x2, y2)
        :param color: Line color
        :param width: Line width in px
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert_vector(pos1, 2)
        assert_vector(pos2, 2)
        color = assert_color(color)
        assert isinstance(width, int) and width >= 1
        length = math.sqrt(math.pow(pos1[0] - pos2[0], 2) + math.pow(pos1[1] - pos2[1], 2))
        assert length > 0, 'line cannot be zero-length'
        return self._add_decor(
            DECORATION_LINE, prev, ((tuple(pos1), tuple(pos2)), color, width, kwargs)
        )

    def add_fill(
        self,
        color: ColorInputType,
        prev: bool = True
    ) -> str:
        """
        Fills the decorator rect object.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param color: Fill color
        :param prev: If ``True`` draw previous the object, else draws post
        :return: ID of the decoration
        """
        return self._add_decor(DECORATION_FILL, prev, assert_color(color))

    def add_hline(
        self,
        x1: NumberType,
        x2: NumberType,
        y: NumberType,
        color: ColorInputType,
        width: int = 1,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a horizontal line.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x1: Horizontal position 1 in px
        :param x2: Horizontal position 2 in px
        :param y: Vertical position in px
        :param color: Line color
        :param width: Line width in px
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert x1 != x2
        return self.add_line((x1, y), (x2, y), color, width, prev, **kwargs)

    def add_vline(
        self,
        x: NumberType,
        y1: NumberType,
        y2: NumberType,
        color: ColorInputType,
        width: int = 1,
        prev: bool = True,
        **kwargs
    ) -> str:
        """
        Adds a vertical line.

        kwargs (Optional)
            - ``use_center_positioning``    (bool) – Uses object center position as *(0, 0)*. ``True`` by default

        :param x: Horizontal position in px
        :param y1: Vertical position 1 in px
        :param y2: Vertical position 2 in px
        :param color: Line color
        :param width: Line width in px
        :param prev: If ``True`` draw previous the object, else draws post
        :param kwargs: Optional keyword arguments
        :return: ID of the decoration
        """
        assert y1 != y2
        return self.add_line((x, y1), (x, y2), color, width, prev, **kwargs)

    def disable(self, decorid: str) -> 'Decorator':
        """
        Disable a certain decoration from ID. Raises ``IndexError`` if decoration was
        not found.

        :param decorid: Decoration ID
        :return: Self reference
        """
        if decorid not in self._decor_enabled.keys():
            raise IndexError(f'decoration<"{decorid}"> was not found')
        self._decor_enabled[decorid] = False
        self.force_cache_update(prev=decorid in self._decor_prev_id)
        return self

    def enable(self, decorid: str) -> 'Decorator':
        """
        Enable a certain decoration from ID. Raises ``IndexError`` if decoration
        was not found.

        :param decorid: Decoration ID
        :return: Self reference
        """
        if decorid not in self._decor_enabled.keys():
            raise IndexError(f'decoration<"{decorid}"> was not found')
        self._decor_enabled[decorid] = True
        self.force_cache_update(prev=decorid in self._decor_prev_id)
        return self

    def is_enabled(self, decorid: str) -> bool:
        """
        Checks if a certain decoration is enabled or not. Raises ``IndexError``
        if decoration was not found.

        :param decorid: Decoration ID
        :return: ``True`` if enabled
        """
        if decorid not in self._decor_enabled.keys():
            raise IndexError(f'decoration<"{decorid}"> was not found')
        return self._decor_enabled[decorid]

    def remove(self, decorid: str) -> 'Decorator':
        """
        Remove a decoration from a given ID. Raises ``IndexError`` if decoration
        was not found.

        :param decorid: Decoration ID
        :return: Self reference
        """
        assert isinstance(decorid, str)
        if decorid in self._coord_cache.keys():
            del self._coord_cache[decorid]
        for p in (DECOR_TYPE_PREV, DECOR_TYPE_POST):
            for d in self._decor[p]:
                if d[1] == decorid:
                    self._decor[p].remove(d)
                    self._cache_needs_update[p] = True
                    if decorid in self._decor_prev_id:
                        self._decor_prev_id.remove(decorid)
                    del self._decor_enabled[decorid]
                    return self
        raise IndexError(f'decoration<"{decorid}"> was not found')

    def remove_all(self, prev: Optional[bool] = None) -> 'Decorator':
        """
        Remove all decorations.

        :param prev: Remove from ``prev`` or ``post``. If ``None`` both are removed
        :return: Self reference
        """
        if prev is None:
            self.remove_all(True)
            self.remove_all(False)
            return self
        p = DECOR_TYPE_PREV if prev else DECOR_TYPE_POST
        self._cache_needs_update[p] = False
        del self._decor[p]
        self._decor[p] = []
        return self

    def _draw_assemble_cache(
        self,
        prev: str,
        deco: List[Tuple[int, str, Any]],
        surface: 'pygame.Surface'
    ) -> None:
        """
        Draw cache, assemble if needed.

        :param prev: Mode
        :param deco: Decoration lists
        :param surface: Source surface to draw from
        """
        if len(deco) == 0:
            return

        w, h = surface.get_size()
        rect = self._obj.get_rect()

        # If it needs update, or the surface size changed, or the rect position changed
        prev_surf_changed = self._cache_last_status[prev][0] != w or \
                            self._cache_last_status[prev][1] != h
        prev_rect_changed = self._cache_last_status[prev][2] != rect.x or \
                            self._cache_last_status[prev][3] != rect.y or \
                            self._cache_last_status[prev][4] != rect.width or \
                            self._cache_last_status[prev][5] != rect.height

        if (
            self._cache_needs_update[prev] or
            prev_surf_changed or
            prev_rect_changed or
            self._cache_surface[prev] is None
        ):
            self._cache_last_status[prev] = (w, h, rect.x, rect.y, rect.width, rect.height)
            del self._cache_surface[prev]
            self._cache_surface[prev] = make_surface(surface.get_width(), surface.get_height())
            self._draw(deco, self._cache_surface[prev])
            self._cache_needs_update[prev] = False

        surface.blit(self._cache_surface[prev], (0, 0))

    def draw_prev(self, surface: 'pygame.Surface') -> 'Decorator':
        """
        Draw prev.

        :param surface: Pygame surface
        :return: Self reference
        """
        if not self.cache:
            self._draw(self._decor[DECOR_TYPE_PREV], surface)
        else:
            self._draw_assemble_cache(DECOR_TYPE_PREV, self._decor[DECOR_TYPE_PREV], surface)
        return self

    def draw_post(self, surface: 'pygame.Surface') -> 'Decorator':
        """
        Draw post.

        :param surface: Pygame surface
        :return: Self reference
        """
        if not self.cache:
            self._draw(self._decor[DECOR_TYPE_POST], surface)
        else:
            self._draw_assemble_cache(DECOR_TYPE_POST, self._decor[DECOR_TYPE_POST], surface)
        return self

    # noinspection PyArgumentList
    def _draw(self, deco: List[Tuple[int, str, Any]], surface: 'pygame.Surface') -> None:
        """
        Draw.

        :param deco: Decoration list
        :param surface: Pygame surface
        """
        if len(deco) == 0:
            return
        rect = self._obj.get_rect()

        for d in deco:
            dtype, decoid, data = d
            if not self._decor_enabled[decoid]:
                continue

            if dtype == DECORATION_POLYGON:
                points, color, filled, width, gfx, kwargs = data
                points = self._update_pos_list(rect, decoid, points, **kwargs)
                if gfx:
                    if filled:
                        gfxdraw.filled_polygon(surface, points, color)
                    else:
                        gfxdraw.polygon(surface, points, color)
                else:
                    pydraw.polygon(surface, color, points, width)

            elif dtype == DECORATION_CIRCLE:
                points, r, color, filled, width, gfx, kwargs = data
                points = self._update_pos_list(rect, decoid, points, **kwargs)
                x, y = points[0]
                if filled:
                    if gfx:
                        gfxdraw.filled_circle(surface, x, y, r, color)
                    else:
                        pydraw.circle(surface, color, (x, y), r)
                else:
                    pydraw.circle(surface, color, (x, y), r, width)

            elif dtype == DECORATION_SURFACE or dtype == DECORATION_BASEIMAGE or dtype == DECORATION_TEXT:
                pos, surf, centered, kwargs = data
                if isinstance(surf, pygame_menu.BaseImage):
                    surf = surf.get_surface(new=False)
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)[0]
                surf_rect = surf.get_rect()
                surf_rect.x += pos[0]
                surf_rect.y += pos[1]
                if centered:
                    surf_rect.x -= int(surf_rect.width / 2)
                    surf_rect.y -= int(surf_rect.height / 2)
                surface.blit(surf, surf_rect)

            elif dtype == DECORATION_ELLIPSE:
                pos, rx, ry, color, filled, kwargs = data
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)[0]
                if filled:
                    gfxdraw.filled_ellipse(surface, pos[0], pos[1], rx, ry, color)
                else:
                    gfxdraw.ellipse(surface, pos[0], pos[1], rx, ry, color)

            elif dtype == DECORATION_CALLABLE:
                data(surface, self._obj)

            elif dtype == DECORATION_CALLABLE_NO_ARGS:
                data()

            elif dtype == DECORATION_TEXTURE_POLYGON:
                pos, texture, tx, ty, kwargs = data
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)
                if isinstance(texture, pygame_menu.BaseImage):
                    texture = texture.get_surface()
                gfxdraw.textured_polygon(surface, pos, texture, tx, ty)

            elif dtype == DECORATION_ARC:
                points, r, ia, fa, color, width, gfx, kwargs = data
                points = self._update_pos_list(rect, decoid, points, **kwargs)
                x, y = points[0]
                rect_arc = pygame.Rect(x - r, y - r, x + 2 * r, y + 2 * r)
                if gfx:
                    gfxdraw.arc(surface, x, y, r, ia, fa, color)
                else:
                    pydraw.arc(surface, color, rect_arc, ia / (2 * pi), fa / (2 * pi), width)

            elif dtype == DECORATION_PIE:
                points, r, ia, fa, color, kwargs = data
                points = self._update_pos_list(rect, decoid, points, **kwargs)
                x, y = points[0]
                gfxdraw.pie(surface, x, y, r, ia, fa, color)

            elif dtype == DECORATION_BEZIER:
                points, color, steps, kwargs = data
                points = self._update_pos_list(rect, decoid, points, **kwargs)
                gfxdraw.bezier(surface, points, steps, color)

            elif dtype == DECORATION_FILL:
                surface.fill(data, rect)

            elif dtype == DECORATION_RECT:
                d_rect: 'pygame.Rect'
                pos, d_rect, color, width, kwargs = data
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)[0]
                d_rect = d_rect.copy()
                d_rect.x += pos[0]
                d_rect.y += pos[1]
                pygame.draw.rect(surface, color, d_rect, width)

            elif dtype == DECORATION_PIXEL:
                pos, color, kwargs = data
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)[0]
                gfxdraw.pixel(surface, pos[0], pos[1], color)

            elif dtype == DECORATION_LINE:
                pos, color, width, kwargs = data
                pos = self._update_pos_list(rect, decoid, pos, **kwargs)
                pydraw.line(surface, color, pos[0], pos[1], width)

            else:
                raise ValueError('unknown decoration type')

    def _update_pos_list(
        self,
        rect: 'pygame.Rect',
        decoid: str,
        pos: Union[Tuple2NumberType, Tuple[Tuple2NumberType, ...]],  # only (x, y) or ((x1,y1), ...
        use_center_positioning=True
    ) -> Union[Tuple[Tuple2IntType, ...], Tuple2IntType]:
        """
        Updates position list based on rect center. If position of the rect changes,
        update the coords.

        :param rect: Object precomputed rect
        :param decoid: Decoration id
        :param pos: Original position tuple of the decoration
        :param use_center_positioning: If ``True`` use *(0, 0)* as the object center
        :return: Position list updated to
        """
        if not use_center_positioning:
            return tuple(pos)

        cx, cy = rect.centerx, rect.centery  # Center position

        # Position of the rect has not changed and exists
        decoid_exists = False
        try:
            decoid_exists = self._coord_cache[decoid] is not None
        except KeyError:
            pass
        if decoid_exists and self._coord_cache[decoid][0] == cx and self._coord_cache[decoid][1] == cy:
            return self._coord_cache[decoid][2]

        # Update the position
        new_pos = []
        for p in pos:
            new_pos.append((int(p[0] + cx), int(p[1] + cy)))
        new_pos = tuple(new_pos)
        self._coord_cache[decoid] = (cx, cy, new_pos)

        return new_pos


class _DecoratorCopyException(Exception):
    """
    If user tries to copy a Decorator.
    """
    pass



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

PYINSTALLER
Module for pyinstaller.
"""

__all__ = ['get_hook_dirs']

import os
from typing import List


def get_hook_dirs() -> List[str]:
    """
    Return hook dirs to PyInstaller.

    :return: Hook dir list
    """
    return [os.path.dirname(__file__)]



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - SCROLL MENU
Shows scrolling in menu.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

from typing import Any
from functools import partial

FPS = 30
WINDOW_SIZE = (800, 600)


def on_button_click(value: str, text: Any = None) -> None:
    """
    Button event on menus.

    :param value: Button value
    :param text: Button text
    """
    if not text:
        print(f'Hello from {value}')
    else:
        print(f'Hello from {text} with {value}')


def paint_background(surface: 'pygame.Surface') -> None:
    """
    Paints a given surface with background color.

    :param surface: Pygame surface
    """
    surface.fill((128, 230, 198))


def make_long_menu() -> 'pygame_menu.Menu':
    """
    Create a long scrolling menu.

    :return: Menu
    """
    theme_menu = pygame_menu.themes.THEME_BLUE.copy()
    theme_menu.scrollbar_cursor = pygame_menu.locals.CURSOR_HAND

    # Main menu, pauses execution of the application
    menu = pygame_menu.Menu(
        height=400,
        onclose=pygame_menu.events.EXIT,
        theme=theme_menu,
        title='Main Menu',
        width=600
    )

    menu_sub = pygame_menu.Menu(
        columns=4,
        height=400,
        onclose=pygame_menu.events.EXIT,
        rows=3,
        theme=pygame_menu.themes.THEME_GREEN,
        title='Menu with columns',
        width=600
    )

    menu_contributors = pygame_menu.Menu(
        height=400,
        onclose=pygame_menu.events.EXIT,
        theme=pygame_menu.themes.THEME_SOLARIZED,
        title='Contributors',
        width=600
    )

    # Add table to contributors
    table_contrib = menu_contributors.add.table()
    table_contrib.default_cell_padding = 5
    table_contrib.default_row_background_color = 'white'
    bold_font = pygame_menu.font.FONT_OPEN_SANS_BOLD
    table_contrib.add_row(['N°', 'Github User'], cell_font=bold_font)
    for i in range(len(pygame_menu.__contributors__)):
        table_contrib.add_row([i + 1, pygame_menu.__contributors__[i]], cell_font=bold_font if i == 0 else None)

    table_contrib.update_cell_style(-1, -1, font_size=15)  # Update all column/row
    table_contrib.update_cell_style(1, [2, -1], font=pygame_menu.font.FONT_OPEN_SANS_ITALIC)

    menu_text = pygame_menu.Menu(
        height=400,
        onclose=pygame_menu.events.EXIT,
        theme=pygame_menu.themes.THEME_DARK,
        title='Text with scroll',
        width=600
    )

    menu.add.button('Rows and Columns', menu_sub)
    menu.add.button('Text scrolled', menu_text)
    menu.add.button('Pygame-menu contributors', menu_contributors)
    menu.add.vertical_margin(20)  # Adds margin

    label1 = 'Button n°{}'
    label2 = 'Text n°{}: '
    for i in range(1, 20):
        if i % 2 == 0:
            menu.add.button(label1.format(i),
                            on_button_click,
                            f'Button n°{i}')
        else:
            menu.add.text_input(label2.format(i),
                                onchange=on_button_click,
                                text=f'Text n°{i}')
    menu.add.button('Exit', pygame_menu.events.EXIT)

    label = 'Button n°{}'
    for i in range(1, 11):
        # Test large button
        if i == 5:
            txt = 'This is a very long button!'
        else:
            txt = label.format(100 * i)
        menu_sub.add.button(txt, on_button_click, 100 * i)
    menu_sub.add.button('Back', pygame_menu.events.BACK)

    # noinspection SpellCheckingInspection
    menu_text.add.label(
        'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod '
        'tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim '
        'veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea '
        'commodo consequat. Duis aute irure dolor in reprehenderit in voluptate '
        'velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat '
        'cupidatat non proident, sunt in culpa qui officia deserunt mollit anim '
        'id est laborum.',
        max_char=33,
        align=pygame_menu.locals.ALIGN_LEFT,
        margin=(0, -1)
    )
    return menu


def main(test: bool = False) -> None:
    """
    Main function.

    :param test: Indicate function is being tested
    """
    screen = create_example_window('Example - Scrolling Menu', WINDOW_SIZE)

    clock = pygame.time.Clock()
    menu = make_long_menu()

    # -------------------------------------------------------------------------
    # Main loop
    # -------------------------------------------------------------------------
    while True:

        # Tick
        clock.tick(FPS)

        # Paint background
        paint_background(screen)

        # Execute main from principal menu if is enabled
        menu.mainloop(
            surface=screen,
            bgfun=partial(paint_background, screen),
            disable_loop=test,
            fps_limit=FPS
        )

        # Update surface
        pygame.display.flip()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

LEFT ARROW CLASS
Selector with a left arrow on the item.
"""

__all__ = ['LeftArrowSelection']

import pygame
import pygame_menu

from pygame_menu.widgets.selection.arrow_selection import ArrowSelection

from pygame_menu._types import Tuple2IntType, NumberType, NumberInstance


class LeftArrowSelection(ArrowSelection):
    """
    Widget selection left arrow class.
    Creates an arrow to the left of the selected Menu item.

    :param arrow_size: Size of arrow on x-axis and y-axis (width, height) in px
    :param arrow_right_margin: Distance from the arrow to the widget (px)
    :param arrow_vertical_offset: Vertical offset of the arrow (px)
    :param blink_ms: Milliseconds between each blink; if ``0`` blinking is disabled
    """
    _arrow_right_margin: int

    def __init__(
        self,
        arrow_size: Tuple2IntType = (10, 15),
        arrow_right_margin: int = 5,
        arrow_vertical_offset: int = 0,
        blink_ms: NumberType = 0
    ) -> None:
        assert isinstance(arrow_right_margin, NumberInstance)
        assert arrow_right_margin >= 0, 'margin cannot be negative'

        super(LeftArrowSelection, self).__init__(
            margin_left=arrow_size[0] + arrow_right_margin,
            margin_right=0,
            margin_top=0,
            margin_bottom=0,
            arrow_vertical_offset=arrow_vertical_offset,
            blink_ms=blink_ms
        )

        self._arrow_right_margin = arrow_right_margin

    # noinspection PyMissingOrEmptyDocstring
    def draw(self, surface: 'pygame.Surface', widget: 'pygame_menu.widgets.Widget') -> 'LeftArrowSelection':
        # A
        #   \B        widget
        # C /
        #     <------>
        #      margin
        rect = widget.get_rect()
        a = (rect.topleft[0] - self._arrow_size[0] - self._arrow_right_margin,
             int(rect.midleft[1] - self._arrow_size[1] / 2 + self._arrow_vertical_offset))
        b = (rect.midleft[0] - self._arrow_right_margin,
             rect.midleft[1] + self._arrow_vertical_offset)
        c = (rect.bottomleft[0] - self._arrow_size[0] - self._arrow_right_margin,
             int(rect.midleft[1] + self._arrow_size[1] / 2 + self._arrow_vertical_offset))
        super(LeftArrowSelection, self)._draw_arrow(surface, widget, a, b, c)
        return self



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - GAME SELECTOR
Game with 3 difficulty options.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

from random import randrange
from typing import Tuple, Any, Optional, List

# Constants and global variables
ABOUT = [f'pygame-menu {pygame_menu.__version__}',
         f'Author: {pygame_menu.__author__}',
         f'Email: {pygame_menu.__email__}']
DIFFICULTY = ['EASY']
FPS = 60
WINDOW_SIZE = (640, 480)

clock: Optional['pygame.time.Clock'] = None
main_menu: Optional['pygame_menu.Menu'] = None
surface: Optional['pygame.Surface'] = None


def change_difficulty(value: Tuple[Any, int], difficulty: str) -> None:
    """
    Change difficulty of the game.

    :param value: Tuple containing the data of the selected object
    :param difficulty: Optional parameter passed as argument to add_selector
    """
    selected, index = value
    print(f'Selected difficulty: "{selected}" ({difficulty}) at index {index}')
    DIFFICULTY[0] = difficulty


def random_color() -> Tuple[int, int, int]:
    """
    Return a random color.

    :return: Color tuple
    """
    return randrange(0, 255), randrange(0, 255), randrange(0, 255)


def play_function(difficulty: List, font: 'pygame.font.Font', test: bool = False) -> None:
    """
    Main game function.

    :param difficulty: Difficulty of the game
    :param font: Pygame font
    :param test: Test method, if ``True`` only one loop is allowed
    """
    assert isinstance(difficulty, list)
    difficulty = difficulty[0]
    assert isinstance(difficulty, str)

    # Define globals
    global main_menu
    global clock

    if difficulty == 'EASY':
        f = font.render('Playing as a baby (easy)', True, (255, 255, 255))
    elif difficulty == 'MEDIUM':
        f = font.render('Playing as a kid (medium)', True, (255, 255, 255))
    elif difficulty == 'HARD':
        f = font.render('Playing as a champion (hard)', True, (255, 255, 255))
    else:
        raise ValueError(f'unknown difficulty {difficulty}')
    f_esc = font.render('Press ESC to open the menu', True, (255, 255, 255))

    # Draw random color and text
    bg_color = random_color()

    # Reset main menu and disable
    # You also can set another menu, like a 'pause menu', or just use the same
    # main_menu as the menu that will check all your input.
    main_menu.disable()
    main_menu.full_reset()

    frame = 0

    while True:

        # noinspection PyUnresolvedReferences
        clock.tick(60)
        frame += 1

        # Application events
        events = pygame.event.get()
        for e in events:
            if e.type == pygame.QUIT:
                exit()
            elif e.type == pygame.KEYDOWN:
                if e.key == pygame.K_ESCAPE:
                    main_menu.enable()

                    # Quit this function, then skip to loop of main-menu on line 221
                    return

        # Pass events to main_menu
        if main_menu.is_enabled():
            main_menu.update(events)

        # Continue playing
        surface.fill(bg_color)
        surface.blit(f, (int((WINDOW_SIZE[0] - f.get_width()) / 2), int(WINDOW_SIZE[1] / 2 - f.get_height())))
        surface.blit(f_esc, (int((WINDOW_SIZE[0] - f_esc.get_width()) / 2), int(WINDOW_SIZE[1] / 2 + f_esc.get_height())))
        pygame.display.flip()

        # If test returns
        if test and frame == 2:
            break


def main_background() -> None:
    """
    Function used by menus, draw on background while menu is active.
    """
    global surface
    surface.fill((128, 0, 128))


def main(test: bool = False) -> None:
    """
    Main program.

    :param test: Indicate function is being tested
    """

    # -------------------------------------------------------------------------
    # Globals
    # -------------------------------------------------------------------------
    global clock
    global main_menu
    global surface

    # -------------------------------------------------------------------------
    # Create window
    # -------------------------------------------------------------------------
    surface = create_example_window('Example - Game Selector', WINDOW_SIZE)
    clock = pygame.time.Clock()

    # -------------------------------------------------------------------------
    # Create menus: Play Menu
    # -------------------------------------------------------------------------
    play_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.7,
        title='Play Menu',
        width=WINDOW_SIZE[0] * 0.75
    )

    submenu_theme = pygame_menu.themes.THEME_DEFAULT.copy()
    submenu_theme.widget_font_size = 15
    play_submenu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.5,
        theme=submenu_theme,
        title='Submenu',
        width=WINDOW_SIZE[0] * 0.7
    )
    for i in range(30):
        play_submenu.add.button(f'Back {i}', pygame_menu.events.BACK)
    play_submenu.add.button('Return to main menu', pygame_menu.events.RESET)

    play_menu.add.button('Start',  # When pressing return -> play(DIFFICULTY[0], font)
                         play_function,
                         DIFFICULTY,
                         pygame.font.Font(pygame_menu.font.FONT_FRANCHISE, 30))
    play_menu.add.selector('Select difficulty ',
                           [('1 - Easy', 'EASY'),
                            ('2 - Medium', 'MEDIUM'),
                            ('3 - Hard', 'HARD')],
                           onchange=change_difficulty,
                           selector_id='select_difficulty')
    play_menu.add.button('Another menu', play_submenu)
    play_menu.add.button('Return to main menu', pygame_menu.events.BACK)

    # -------------------------------------------------------------------------
    # Create menus:About
    # -------------------------------------------------------------------------
    about_theme = pygame_menu.themes.THEME_DEFAULT.copy()
    about_theme.widget_margin = (0, 0)

    about_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.6,
        theme=about_theme,
        title='About',
        width=WINDOW_SIZE[0] * 0.6
    )

    for m in ABOUT:
        about_menu.add.label(m, align=pygame_menu.locals.ALIGN_LEFT, font_size=20)
    about_menu.add.vertical_margin(30)
    about_menu.add.button('Return to menu', pygame_menu.events.BACK)

    # -------------------------------------------------------------------------
    # Create menus: Main
    # -------------------------------------------------------------------------
    main_theme = pygame_menu.themes.THEME_DEFAULT.copy()

    main_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.6,
        theme=main_theme,
        title='Main Menu',
        width=WINDOW_SIZE[0] * 0.6
    )

    main_menu.add.button('Play', play_menu)
    main_menu.add.button('About', about_menu)
    main_menu.add.button('Quit', pygame_menu.events.EXIT)

    # -------------------------------------------------------------------------
    # Main loop
    # -------------------------------------------------------------------------
    while True:

        # Tick
        clock.tick(FPS)

        # Paint background
        main_background()

        # Application events
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                exit()

        # Main menu
        if main_menu.is_enabled():
            main_menu.mainloop(surface, main_background, disable_loop=test, fps_limit=FPS)

        # Flip surface
        pygame.display.flip()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

MENU WIDGET MANAGER
Easy widget add/remove to Menus.
"""

__all__ = ['WidgetManager']

import pygame_menu

from pygame_menu._base import Base
from pygame_menu.font import assert_font
from pygame_menu.utils import assert_vector, assert_color, assert_cursor, \
    assert_position_vector, warn

# Import widgets
from pygame_menu.widgets.core.widget import Widget, check_widget_mouseleave

from pygame_menu.widgets.widget.button import ButtonManager
from pygame_menu.widgets.widget.colorinput import ColorInputManager
from pygame_menu.widgets.widget.dropselect import DropSelectManager
from pygame_menu.widgets.widget.dropselect_multiple import DropSelectMultipleManager
from pygame_menu.widgets.widget.frame import FrameManager
from pygame_menu.widgets.widget.hmargin import HMarginManager
from pygame_menu.widgets.widget.image import ImageManager
from pygame_menu.widgets.widget.label import LabelManager
from pygame_menu.widgets.widget.menulink import MenuLinkManager
from pygame_menu.widgets.widget.none import NoneWidgetManager
from pygame_menu.widgets.widget.progressbar import ProgressBarManager
from pygame_menu.widgets.widget.rangeslider import RangeSliderManager
from pygame_menu.widgets.widget.selector import SelectorManager
from pygame_menu.widgets.widget.surface import SurfaceWidgetManager
from pygame_menu.widgets.widget.table import TableManager
from pygame_menu.widgets.widget.textinput import TextInputManager
from pygame_menu.widgets.widget.toggleswitch import ToggleSwitchManager
from pygame_menu.widgets.widget.vfill import VFillManager
from pygame_menu.widgets.widget.vmargin import VMarginManager

from pygame_menu._types import Any, Dict, PaddingInstance


# noinspection PyProtectedMember
class WidgetManager(
    Base,
    ButtonManager,
    ColorInputManager,
    DropSelectManager,
    DropSelectMultipleManager,
    FrameManager,
    HMarginManager,
    ImageManager,
    LabelManager,
    MenuLinkManager,
    NoneWidgetManager,
    ProgressBarManager,
    RangeSliderManager,
    SelectorManager,
    SurfaceWidgetManager,
    TableManager,
    TextInputManager,
    ToggleSwitchManager,
    VFillManager,
    VMarginManager
):
    """
    Add/Remove widgets to the Menu.

    :param menu: Menu reference
    :param verbose: Enable/disable verbose mode (warnings/errors)
    """

    def __init__(self, menu: 'pygame_menu.Menu', verbose: bool = True) -> None:
        super(WidgetManager, self).__init__(object_id=menu.get_id() + '+widget-manager', verbose=verbose)
        self._menu = menu

    @property
    def _theme(self) -> 'pygame_menu.Theme':
        return self._menu.get_theme()

    def _add_submenu(self, menu: 'pygame_menu.Menu', hook: 'Widget') -> None:
        assert isinstance(menu, pygame_menu.Menu)
        assert menu != self._menu, 'submenu cannot point to menu itself'
        assert isinstance(hook, Widget)
        if menu not in self._menu._submenus.keys():
            self._menu._submenus[menu] = []
        assert hook not in self._menu._submenus[menu], \
            f'widget {hook.get_class_id()} already hooks submenu {menu.get_class_id()}'
        self._menu._submenus[menu].append(hook)
        hook._menu_hook = menu

    def _filter_widget_attributes(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        attributes = {}

        # align
        align = kwargs.pop('align', self._theme.widget_alignment)
        assert isinstance(align, str)
        attributes['align'] = align

        # background_color
        background_is_color = False
        background_color = kwargs.pop('background_color', self._theme.widget_background_color)
        if background_color is not None:
            if isinstance(background_color, pygame_menu.BaseImage):
                pass
            else:
                background_color = assert_color(background_color)
                background_is_color = True
        attributes['background_color'] = background_color

        # background_inflate
        background_inflate = kwargs.pop('background_inflate', self._theme.widget_background_inflate)
        if background_inflate == 0:
            background_inflate = (0, 0)
        assert_vector(background_inflate, 2, int)
        assert background_inflate[0] >= 0 and background_inflate[1] >= 0, \
            'both background inflate components must be equal or greater than zero'
        attributes['background_inflate'] = background_inflate

        # border_color
        border_color = kwargs.pop('border_color', self._theme.widget_border_color)
        if border_color is not None:
            border_color = assert_color(border_color)
        attributes['border_color'] = border_color

        # border_inflate
        border_inflate = kwargs.pop('border_inflate', self._theme.widget_border_inflate)
        if border_inflate == 0:
            border_inflate = (0, 0)
        assert_vector(border_inflate, 2, int)
        assert isinstance(border_inflate[0], int) and border_inflate[0] >= 0
        assert isinstance(border_inflate[1], int) and border_inflate[1] >= 0
        attributes['border_inflate'] = border_inflate

        # border_position
        border_position = kwargs.pop('border_position', self._theme.widget_border_position)
        assert_position_vector(border_position)
        attributes['border_position'] = border_position

        # border_width
        border_width = kwargs.pop('border_width', self._theme.widget_border_width)
        assert isinstance(border_width, int) and border_width >= 0
        attributes['border_width'] = border_width

        # cursor
        cursor = kwargs.pop('cursor', self._theme.widget_cursor)
        assert_cursor(cursor)
        attributes['cursor'] = cursor

        # floating status
        float_ = kwargs.pop('float', False)
        assert isinstance(float_, bool)
        attributes['float'] = float_
        float_origin_position = kwargs.pop('float_origin_position', False)
        assert isinstance(float_origin_position, bool)
        attributes['float_origin_position'] = float_origin_position

        # font_antialias
        attributes['font_antialias'] = self._theme.widget_font_antialias

        # font_background_color
        font_background_color = kwargs.pop('font_background_color',
                                           self._theme.widget_font_background_color)
        if (
            font_background_color is None and
            self._theme.widget_font_background_color_from_menu and
            not background_is_color
        ):
            if not isinstance(self._theme.background_color, pygame_menu.BaseImage):
                font_background_color = assert_color(self._theme.background_color)
        attributes['font_background_color'] = font_background_color

        # font_color
        font_color = kwargs.pop('font_color', self._theme.widget_font_color)
        attributes['font_color'] = assert_color(font_color)

        # font_name
        font_name = kwargs.pop('font_name', self._theme.widget_font)
        assert_font(font_name)
        attributes['font_name'] = font_name

        # font_shadow
        font_shadow = kwargs.pop('font_shadow', self._theme.widget_font_shadow)
        assert isinstance(font_shadow, bool)
        attributes['font_shadow'] = font_shadow

        # font_shadow_color
        font_shadow_color = kwargs.pop('font_shadow_color', self._theme.widget_font_shadow_color)
        attributes['font_shadow_color'] = assert_color(font_shadow_color)

        # font_shadow_offset
        font_shadow_offset = kwargs.pop('font_shadow_offset', self._theme.widget_font_shadow_offset)
        assert isinstance(font_shadow_offset, int)
        attributes['font_shadow_offset'] = font_shadow_offset

        # font_shadow_position
        font_shadow_position = kwargs.pop('font_shadow_position', self._theme.widget_font_shadow_position)
        assert isinstance(font_shadow_position, str)
        attributes['font_shadow_position'] = font_shadow_position

        # font_size
        font_size = kwargs.pop('font_size', self._theme.widget_font_size)
        assert isinstance(font_size, int)
        assert font_size > 0, 'font size must be greater than zero'
        attributes['font_size'] = font_size

        # margin
        margin = kwargs.pop('margin', self._theme.widget_margin)
        if margin == 0:
            margin = (0, 0)
        assert_vector(margin, 2)
        attributes['margin'] = margin

        # padding
        padding = kwargs.pop('padding', self._theme.widget_padding)
        assert isinstance(padding, PaddingInstance)
        attributes['padding'] = padding

        # readonly_color
        readonly_color = kwargs.pop('readonly_color', self._theme.readonly_color)
        attributes['readonly_color'] = assert_color(readonly_color)

        # readonly_selected_color
        readonly_selected_color = kwargs.pop('readonly_selected_color', self._theme.readonly_selected_color)
        attributes['readonly_selected_color'] = assert_color(readonly_selected_color)

        # selection_color
        selection_color = kwargs.pop('selection_color', self._theme.selection_color)
        attributes['selection_color'] = assert_color(selection_color)

        # selection_effect
        selection_effect = kwargs.pop('selection_effect', self._theme.widget_selection_effect)
        if selection_effect is None:
            selection_effect = pygame_menu.widgets.NoneSelection()
        else:
            selection_effect = selection_effect.copy()
        assert isinstance(selection_effect, pygame_menu.widgets.core.Selection)

        selection_effect.set_color(attributes['selection_color'])
        attributes['selection_effect'] = selection_effect

        # shadow
        attributes['shadow_aa'] = kwargs.pop('shadow_aa', self._theme.widget_shadow_aa)
        attributes['shadow_color'] = kwargs.pop('shadow_color', self._theme.widget_shadow_color)
        attributes['shadow_radius'] = kwargs.pop('shadow_radius', self._theme.widget_shadow_radius)
        attributes['shadow_type'] = kwargs.pop('shadow_type', self._theme.widget_shadow_type)
        attributes['shadow_width'] = kwargs.pop('shadow_width', self._theme.widget_shadow_width)

        # tab_size
        attributes['tab_size'] = kwargs.pop('tab_size', self._theme.widget_tab_size)

        return attributes

    def _configure_widget(self, widget: 'Widget', **kwargs) -> None:
        assert isinstance(widget, Widget)
        widget._verbose = self._verbose

        widget.set_alignment(
            align=kwargs['align']
        )

        widget.set_background_color(
            color=kwargs['background_color'],
            inflate=kwargs['background_inflate']
        )

        widget.set_border(
            color=kwargs['border_color'],
            inflate=kwargs['border_inflate'],
            position=kwargs['border_position'],
            width=kwargs['border_width']
        )

        widget.set_controls(
            joystick=self._menu._joystick,
            keyboard=self._menu._keyboard,
            mouse=self._menu._mouse,
            touchscreen=self._menu._touchscreen
        )

        widget.set_cursor(
            cursor=kwargs['cursor']
        )

        widget.set_float(
            float_status=kwargs['float'],
            origin_position=kwargs['float_origin_position']
        )

        widget.set_font(
            antialias=kwargs['font_antialias'],
            background_color=kwargs['font_background_color'],
            color=kwargs['font_color'],
            font=kwargs['font_name'],
            font_size=kwargs['font_size'],
            readonly_color=kwargs['readonly_color'],
            readonly_selected_color=kwargs['readonly_selected_color'],
            selected_color=kwargs['selection_color']
        )

        widget.set_font_shadow(
            color=kwargs['font_shadow_color'],
            enabled=kwargs['font_shadow'],
            offset=kwargs['font_shadow_offset'],
            position=kwargs['font_shadow_position']
        )

        widget.set_margin(
            x=kwargs['margin'][0],
            y=kwargs['margin'][1]
        )

        widget.set_padding(
            padding=kwargs['padding']
        )

        widget.set_selection_effect(
            selection=kwargs['selection_effect']
        )

        widget.set_tab_size(
            tab_size=kwargs['tab_size']
        )

        widget.shadow(
            aa_amount=kwargs['shadow_aa'],
            color=kwargs['shadow_color'],
            corner_radius=kwargs['shadow_radius'],
            shadow_type=kwargs['shadow_type'],
            shadow_width=kwargs['shadow_width']
        )

        if self._theme.widget_background_inflate_to_selection:
            widget.background_inflate_to_selection_effect()

        widget._update__repr___(self)
        widget._keyboard_ignore_nonphysical = self._menu._keyboard_ignore_nonphysical

        widget.configured = True
        widget._configure()

    @staticmethod
    def _check_kwargs(kwargs: Dict[str, Any]) -> None:
        for invalid_keyword in kwargs.keys():
            raise ValueError(f'widget addition optional parameter kwargs.{invalid_keyword} is not valid')

    def _append_widget(self, widget: 'Widget') -> None:
        assert isinstance(widget, Widget)
        if widget.get_menu() is None:
            widget.set_menu(self._menu)
        assert widget.get_menu() == self._menu, \
            'widget cannot have a different instance of menu'
        self._menu._check_id_duplicated(widget.get_id())

        if widget.get_scrollarea() is None:
            widget.set_scrollarea(self._menu.get_scrollarea())

        # Unselect
        widget.select(False)

        # Append to lists
        self._menu._widgets.append(widget)

        # Update selection index
        if self._menu._index < 0 and widget.is_selectable:
            widget.select()
            self._menu._index = len(self._menu._widgets) - 1

        # Force menu rendering, this checks if the menu overflows or has sizing
        # errors; if added on execution time forces the update of the surface
        self._menu._widgets_surface = None
        try:
            self._menu._render()
        except (pygame_menu.menu._MenuSizingException,
                pygame_menu.menu._MenuWidgetOverflow):
            self._menu.remove_widget(widget)
            raise
        self._menu.render()

        # Sort frame widgets, as render position changes frame position/frame
        if len(self._menu._update_frames) > 0:
            self._menu._update_frames[0]._sort_menu_update_frames()

        # Update widgets
        check_widget_mouseleave()

        # Call event
        widget._append_to_menu()

    # noinspection PyMissingOrEmptyDocstring
    def configure_defaults_widget(self, widget: 'Widget') -> None:
        self._configure_widget(widget, **self._filter_widget_attributes({}))

    def generic_widget(
        self,
        widget: 'Widget',
        configure_defaults: bool = False
    ) -> 'Widget':
        """
        Add generic widget to the Menu.

        .. note::

            The widget should be fully configured by the user: font, padding, etc.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Unintended behaviours may happen while using this method, use only with
            caution; specially while creating nested submenus with buttons.

        :param widget: Widget to be added
        :param configure_defaults: Apply defaults widget configuration (for example, theme)
        :return: The added widget object
        :rtype: :py:class:`pygame_menu.widgets.Widget`
        """
        assert isinstance(widget, Widget)
        if widget.get_menu() is not None:
            raise ValueError('widget to be added is already appended to another Menu')

        # Raise warning if adding button with Menu
        if isinstance(widget, pygame_menu.widgets.Button) and widget.to_menu:
            if self._verbose:
                warn(
                    'prefer adding submenus using add_button method instead, '
                    'unintended behaviours may occur'
                )

        # Configure widget
        if configure_defaults:
            self.configure_defaults_widget(widget)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

LABEL
Label class, adds a simple text to the Menu.
"""

__all__ = [
    'Label',
    'LabelManager'
]

import pygame
import pygame_menu
import textwrap
import time

from abc import ABC
from pygame_menu.utils import assert_color, warn, uuid4, make_surface
from pygame_menu.widgets.core.widget import Widget, AbstractWidgetManager

from pygame_menu._types import Any, CallbackType, List, Union, Tuple, Optional, \
    ColorType, ColorInputType, EventVectorType, Callable

LabelTitleGeneratorType = Optional[Callable[[], str]]


# noinspection PyMissingOrEmptyDocstring
class Label(Widget):
    """
    Label widget.

    .. note::

        Label accepts all transformations.

    :param title: Label title/text
    :param label_id: Label ID
    :param onselect: Function when selecting the label widget
    :param wordwrap: Wraps label if newline is found on widget
    :param leading: Font leading for ``wordwrap``. If ``None`` retrieves from widget font
    :param max_nlines: Number of maximum lines for ``wordwrap``. If ``None`` the number is dynamically computed. If exceded, ``label.get_overflow_lines()`` will return the lines not displayed
    """
    _last_underline: List[Union[str, Optional[Tuple[ColorType, int, int]]]]
    _leading: Optional[int]
    _lines: List[str]
    _max_nlines: Optional[int]
    _overflow_lines: List[str]  # Store how many lines are overflowed starting end
    _title_generator: LabelTitleGeneratorType
    _wordwrap: bool

    def __init__(
        self,
        title: Any,
        label_id: str = '',
        onselect: CallbackType = None,
        wordwrap: bool = False,
        leading: Optional[int] = None,
        max_nlines: Optional[int] = None
    ) -> None:
        assert isinstance(leading, (type(None), int))
        assert isinstance(max_nlines, (type(None), int))
        super(Label, self).__init__(
            title=title,
            onselect=onselect,
            widget_id=label_id
        )
        self._last_underline = ['', None]  # deco id, (color, offset, width)
        self._leading = leading
        self._lines = []  # Lines of text displayed
        self._max_nlines = max_nlines
        self._overflow_lines = []
        self._title_generator = None
        self._wordwrap = wordwrap

    def add_underline(
        self,
        color: ColorInputType,
        offset: int,
        width: int,
        force_render: bool = False
    ) -> 'Label':
        """
        Adds an underline to text. This is added if widget is rendered. Underline
        is only enabled for non wordwrap label.

        :param color: Underline color
        :param offset: Underline offset
        :param width: Underline width
        :param force_render: If ``True`` force widget render after addition
        :return: Self reference
        """
        assert not self._wordwrap, 'underline is not enabled for wordwrap is active'
        color = assert_color(color)
        assert isinstance(offset, int)
        assert isinstance(width, int) and width > 0
        self._last_underline[1] = (color, offset, width)
        if force_render:
            self._force_render()
        return self

    def remove_underline(self) -> 'Label':
        """
        Remove underline of the label.

        :return: Self reference
        """
        assert not self._wordwrap, 'underline is not enabled for wordwrap is active'
        if self._last_underline[0] != '':
            self._decorator.remove(self._last_underline[0])
            self._last_underline[0] = ''
        return self

    def _apply_font(self) -> None:
        pass

    def _draw(self, surface: 'pygame.Surface') -> None:
        # The minimal width of any surface is 1px, so the background will be a line
        if self._title == '':
            return
        surface.blit(self._surface, self._rect.topleft)

    def set_title_generator(self, generator: LabelTitleGeneratorType) -> 'Label':
        """
        Set a title generator. This function is executed each time the label updates,
        returning a new title (string) which replaces the current label title.

        The generator does not take any input as argument.

        :param generator: Function which generates a new text status
        :return: Self reference
        """
        if generator is not None:
            assert callable(generator)
        self._title_generator = generator

        # Update update widgets
        menu_update_widgets = self._get_menu_update_widgets()
        if generator is None and self in menu_update_widgets:
            menu_update_widgets.remove(self)
        if generator is not None and self not in menu_update_widgets:
            menu_update_widgets.append(self)
        return self

    def set_title(self, title: str) -> 'Label':
        super(Label, self).set_title(title)
        if self._title_generator is not None:
            if self._verbose:
                warn(
                    f'{self.get_class_id()} title generator is not None, thus, the new'
                    f' title "{title}" will be overridden after next update'
                )
        return self

    def _get_leading(self) -> int:
        """
        Computes the font leading.

        :return: Leading
        """
        assert self._font
        return (
            self._font.get_linesize()
            if self._leading is None
            else self._leading
        )

    def get_lines(self) -> List[str]:
        """
        Return the lines of text displayed. Each new line belongs to an item on list.

        :return: List of displayed lines
        """
        return self._lines

    @staticmethod
    def _wordwrap_line(
        line: str,
        font: pygame.font.Font,
        max_width: int,
        tab_size: int,
    ) -> List[str]:
        """
        Wordwraps line.

        :param line: Line
        :param font: Font
        :param max_width: Max width
        :param tab_size: Tab size
        :return: List of strings
        """
        final_lines = []
        words = line.split(' ')
        i, current_line = 0, ''

        while True:
            split_line = False
            for i, _ in enumerate(words):
                current_line = ' '.join(words[:i + 1])
                current_line = current_line.replace('\t', ' ' * tab_size)
                current_line_size = font.size(current_line)
                if current_line_size[0] > max_width:
                    split_line = True
                    break

            if split_line:
                i = i if i > 0 else 1
                final_lines.append(' '.join(words[:i]))
                words = words[i:]
            else:
                final_lines.append(current_line)
                break

        return final_lines

    def _get_max_container_width(self) -> int:
        """
        Return the maximum label container width. It can be the column width,
        menu width or frame width if horizontal.

        :return: Container width
        """
        menu = self._menu
        if menu is None:
            return 0
        try:
            # noinspection PyProtectedMember
            max_width = menu._column_widths[self.get_col_row_index()[0]]
        except IndexError:
            max_width = menu.get_width(inner=True)
        return max_width - self._padding[1] - self._padding[3] - self._selection_effect.get_width()

    def get_overflow_lines(self) -> List[str]:
        """
        Return the overflow lines ir ``wordwrap`` is active and ``max_nlines`` is set.

        :return: Lines not displayed
        """
        assert self._wordwrap, 'wordwrap must be enabled'
        assert isinstance(self._max_nlines, int), 'max_nlines must be defined'
        return self._overflow_lines

    def _render(self) -> Optional[bool]:
        if not self._render_hash_changed(self._title, self._font_color, self._visible, self._menu, self._font,
                                         self._last_underline[1], self._padding, self._selection_effect.get_width()):
            return True
        self._lines = []

        # Generate surface
        if not self._wordwrap:
            self._surface = self._render_string(self._title, self._font_color)
            self._lines.append(self._title)

        else:
            self._overflow_lines = []
            if self._font is None or self._menu is None:
                self._surface = make_surface(0, 0, alpha=True)
            else:
                lines = self._title.split('\n')
                lines = sum(
                    (
                        self._wordwrap_line(
                            line=line,
                            font=self._font,
                            max_width=self._get_max_container_width(),
                            tab_size=self._tab_size
                        )
                        for line in lines
                    ),
                    []
                )
                num_lines = len(lines)
                if isinstance(self._max_nlines, int):
                    if num_lines > self._max_nlines:
                        for j in range(num_lines - self._max_nlines):
                            self._overflow_lines.append(lines[num_lines - j - 1])
                    num_lines = min(num_lines, self._max_nlines)

                self._surface = make_surface(
                    max(self._font.size(line)[0] for line in lines),
                    num_lines * self._get_leading(),
                    alpha=True
                )

                for n_line, line in enumerate(lines):
                    line_surface = self._render_string(line, self._font_color)
                    self._surface.blit(
                        line_surface,
                        pygame.Rect(
                            0,
                            n_line * self._get_leading(),
                            self._rect.width,
                            self._rect.height
                        )
                    )
                    self._lines.append(line)
                    if n_line + 1 == num_lines:
                        break

        # Update rect object
        self._apply_transforms()
        self._rect.width, self._rect.height = self._surface.get_size()

        # Apply underline
        if not self._wordwrap:
            self.remove_underline()
            if self._last_underline[1] is not None:
                w = self._surface.get_width()
                h = self._surface.get_height()
                color, offset, width = self._last_underline[1]
                if w > 0 and h > 0:
                    self._last_underline[0] = self._decorator.add_line(
                        pos1=(-w / 2, h / 2 + offset),
                        pos2=(w / 2, h / 2 + offset),
                        color=color,
                        width=width
                    )

        self.force_menu_surface_update()

    def update(self, events: EventVectorType) -> bool:
        # If generator is not None
        if self._title_generator is not None:
            gen_title = self._title_generator()
            assert isinstance(gen_title, str), \
                f'object generated by the title generator ({gen_title}) is not string-type'
            self._title = gen_title
            self._render()
        self.apply_update_callbacks(events)
        for event in events:
            if self._check_mouseover(event):
                break
        return False


class LabelManager(AbstractWidgetManager, ABC):
    """
    Label manager.
    """

    # noinspection PyProtectedMember
    def label(
        self,
        title: Any,
        label_id: str = '',
        max_char: int = 0,
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        selectable: bool = False,
        wordwrap: bool = False,
        **kwargs
    ) -> Union['pygame_menu.widgets.Label', List['pygame_menu.widgets.Label']]:
        """
        Add a simple text to the Menu.

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``leading``                       (int) - Font leading for ``wordwrap``. If ``None`` retrieves from widget font
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``max_nlines``                    (int) - Number of maximum lines for ``wordwrap``. If ``None`` the number is dynamically computed. If exceded, ``label.get_overflow_lines()`` will return the lines not displayed
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect. Applied only if ``selectable`` is ``True``
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character
            - ``underline_color``               (tuple, list, str, int, :py:class:`pygame.Color`, None) – Color of the underline. If ``None`` use the same color of the text
            - ``underline_offset``              (int) – Vertical offset in px. ``2`` by default
            - ``underline_width``               (int) – Underline width in px. ``2`` by default
            - ``underline``                     (bool) – Enables text underline, using a properly placed decoration. ``False`` by default

        .. note::

            All theme-related optional kwargs use the default Menu theme if not defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param title: Text to be displayed
        :param label_id: ID of the label
        :param max_char: Split the title in several labels if the string length exceeds ``max_char``; ``0``: don't split, ``-1``: split to Menu width
        :param onselect: Callback executed when selecting the widget; only executed if ``selectable`` is ``True``
        :param selectable: Label accepts user selection; useful to move along the Menu using label selection
        :param wordwrap: Wraps label if newline is found on widget. If ``False`` the manager splits the string and creates a list of widgets, else, the widget itself splits and updates the height
        :param kwargs: Optional keyword arguments
        :return: Widget object, or List of widgets if the text overflows
        :rtype: :py:class:`pygame_menu.widgets.Label`, :py:class:`typing.List` [:py:class:`pygame_menu.widgets.Label`]
        """
        assert isinstance(label_id, str)
        assert isinstance(max_char, int)
        assert isinstance(selectable, bool)
        assert max_char >= -1

        title = str(title)
        if len(label_id) == 0:
            label_id = uuid4()

        # If newline detected, split in two new lines
        if '\n' in title and not wordwrap:
            title = title.split('\n')
            widgets = []
            for t in title:
                wig = self.label(
                    title=t,
                    label_id=label_id + '+' + str(len(widgets) + 1),
                    max_char=max_char,
                    onselect=onselect,
                    selectable=selectable,
                    **kwargs
                )
                if isinstance(wig, list):
                    for w in wig:
                        widgets.append(w)
                else:
                    widgets.append(wig)
            return widgets

        # Wrap text to Menu width (imply additional calls to render functions)
        if max_char < 0:
            if len(title) > 0:
                dummy_attrs = self._filter_widget_attributes(kwargs.copy())
                dummy = pygame_menu.widgets.Label(title=title)
                self._configure_widget(dummy, **dummy_attrs)
                max_char = int(1.0 * self._menu.get_width(inner=True) * len(title) / dummy.get_width())
            else:
                max_char = 0

        leading = kwargs.pop('leading', None)
        max_nlines = kwargs.pop('max_nlines', None)

        # If no overflow
        if len(title) <= max_char or max_char == 0 or wordwrap:
            attributes = self._filter_widget_attributes(kwargs)

            # Filter additional parameters
            underline = kwargs.pop('underline', False)
            underline_color = kwargs.pop('underline_color', attributes['font_color'])
            underline_offset = kwargs.pop('underline_offset', 1)
            underline_width = kwargs.pop('underline_width', 1)

            widget = Label(
                label_id=label_id,
                onselect=onselect,
                title=title,
                wordwrap=wordwrap and not underline,
                leading=leading,
                max_nlines=max_nlines
            )
            widget.is_selectable = selectable
            self._check_kwargs(kwargs)
            self._configure_widget(widget=widget, **attributes)

            if underline:
                widget.add_underline(underline_color, underline_offset, underline_width)

            self._append_widget(widget)

        else:
            self._menu._check_id_duplicated(label_id)  # Before adding + LEN
            widget = []
            for line in textwrap.wrap(title, max_char):
                widget.append(
                    self.label(
                        title=line,
                        label_id=label_id + '+' + str(len(widget) + 1),
                        max_char=max_char,
                        onselect=onselect,
                        selectable=selectable,
                        wordwrap=wordwrap,
                        leading=leading,
                        max_nlines=max_nlines,
                        **kwargs
                    )
                )

        return widget

    def clock(
        self,
        clock_format: str = '%Y/%m/%d %H:%M:%S',
        clock_id: str = '',
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        selectable: bool = False,
        title_format: str = '{0}',
        wordwrap: bool = False,
        **kwargs
    ) -> 'pygame_menu.widgets.Label':
        """
        Add a clock label to the Menu. This creates a Label with a text generator
        that request a string from ``time.strftime`` module using ``clock_format``.

        Commonly used format codes:
            - **%Y**    – Year with century as a decimal number
            - **%m**    – Month as a decimal number [01, 12]
            - **%d**    – Day of the month as a decimal number [01, 31]
            - **%H**    – Hour (24-hour clock) as a decimal number [00, 23]
            - **%M**    – Minute as a decimal number [00, 59]
            - **%S**    – Second as a decimal number [00, 61]
            - **%z**    – Time zone offset from UTC
            - **%a**    – Locale's abbreviated weekday name
            - **%A**    – Locale's full weekday name
            - **%b**    – Locale's abbreviated month name
            - **%B**    – Locale's full month name
            - **%c**    – Locale's appropriate date and time representation
            - **%I**    – Hour (12-hour clock) as a decimal number [01, 12]
            - **%p**    – Locale's equivalent of either AM or PM

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int,  :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``leading``                       (int) - Font leading for ``wordwrap``. If ``None`` retrieves from widget font
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``max_nlines``                    (int) - Number of maximum lines for ``wordwrap``. If ``None`` the number is dynamically computed. If exceded, ``label.get_overflow_lines()`` will return the lines not displayed
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect. Applied only if ``selectable`` is ``True``
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character
            - ``underline_color``               (tuple, list, str, int, :py:class:`pygame.Color`, None) – Color of the underline. If ``None`` use the same color of the text
            - ``underline_offset``              (int) – Vertical offset in px. ``2`` by default
            - ``underline_width``               (int) – Underline width in px. ``2`` by default
            - ``underline``                     (bool) – Enables text underline, using a properly placed decoration. ``False`` by default

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param clock_format: Format of clock used by ``time.strftime``
        :param clock_id: ID of the clock
        :param onselect: Callback executed when selecting the widget; only executed if ``selectable`` is ``True``
        :param selectable: Label accepts user selection; useful to move along the Menu using label selection
        :param title_format: Title format which accepts ``{0}`` as the string from ``time.strftime``, for example, ``'My Clock {0}'`` can be a title format
        :param wordwrap: Wraps label if newline is found on widget. If ``False`` the manager splits the string and creates a list of widgets, else, the widget itself splits and updates the height
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Label`
        """
        label = self.label(
            title='',
            label_id=clock_id,
            onselect=onselect,
            selectable=selectable,
            wordwrap=wordwrap,
            **kwargs
        )

        assert isinstance(title_format, str) and '{0}' in title_format
        assert not isinstance(label, list)
        label.set_title_generator(lambda: title_format.format(time.strftime(clock_format)))
        label.update([])

        return label



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

BASE
Base object. Provides common methods used by all library objects.
"""

__all__ = ['Base']

from pygame_menu.utils import uuid4

from pygame_menu._types import Dict, Any, NumberInstance, NumberType, Optional


class Base(object):
    """
    Base object.
    """
    _attributes: Optional[Dict[str, Any]]
    _class_id__repr__: bool
    _id: str
    _id__repr__: bool
    _verbose: bool

    def __init__(self, object_id: str, verbose: bool = True) -> None:
        """
        Base constructor.

        :param object_id: Object ID
        :param verbose: Enable verbose mode (errors/warnings)
        """
        assert isinstance(object_id, str)
        assert isinstance(verbose, bool)
        if len(object_id) == 0:
            object_id = uuid4()
        self._attributes = None
        self._class_id__repr__ = False  # If True, repr/str of the object is class id
        self._id = object_id
        self._id__repr__ = False  # If True, repr/str of the object adds object id
        self._verbose = verbose

    def __repr__(self) -> str:
        """
        Repr print of object.

        :return: Object str status
        """
        sup_repr = super(Base, self).__repr__()
        assert not (self._class_id__repr__ and self._id__repr__), \
            'class id and id __repr__ cannot be True at the same time'
        if self._class_id__repr__:
            return self.get_class_id()
        if self._id__repr__:
            return sup_repr.replace(' object at ', f'["{self.get_id()}"] object at ')
        return sup_repr

    def _update__repr___(self, obj: 'Base') -> None:
        """
        Update __repr__ from other Base object.

        :param obj: External base object to copy from
        """
        self._class_id__repr__ = obj._class_id__repr__
        self._id__repr__ = obj._id__repr__

    def set_attribute(self, key: str, value: Any = None) -> 'Base':
        """
        Set an attribute.

        :param key: Key of the attribute
        :param value: Value of the attribute
        :return: Self reference
        """
        assert isinstance(key, str)
        if self._attributes is None:
            self._attributes = {}
        self._attributes[key] = value
        return self

    def get_counter_attribute(self, key: str, incr: Any = 0, default: Any = 0) -> NumberType:
        """
        Get counter attribute.

        :param key: Key of the attribute
        :param incr: Increase value
        :param default: Default vale to start with, by default it's zero
        :return: New increase value
        """
        if not isinstance(incr, NumberInstance):
            incr = float(incr)
        if not isinstance(default, NumberInstance):
            if isinstance(incr, float):
                default = float(default)
            else:
                try:
                    default = int(default)
                except ValueError:
                    default = float(default)
        if not self.has_attribute(key):
            self.set_attribute(key, default + incr)
            return default + incr
        new = self.get_attribute(key) + incr
        self.set_attribute(key, new)
        return new

    def get_attribute(self, key: str, default: Any = None) -> Any:
        """
        Get an attribute value.

        :param key: Key of the attribute
        :param default: Value if it does not exist
        :return: Attribute data
        """
        assert isinstance(key, str)
        if not self.has_attribute(key):
            return default
        return self._attributes[key]

    def has_attribute(self, key: str) -> bool:
        """
        Return ``True`` if the object has the given attribute.

        :param key: Key of the attribute
        :return: ``True`` if exists
        """
        assert isinstance(key, str)
        if self._attributes is None:
            return False
        return key in self._attributes.keys()

    def remove_attribute(self, key: str) -> 'Base':
        """
        Removes the given attribute from the object. Throws ``IndexError`` if
        the given key does not exist.

        :param key: Key of the attribute
        :return: Self reference
        """
        if not self.has_attribute(key):
            raise IndexError(f'attribute "{key}" does not exists on object')
        del self._attributes[key]
        return self

    def get_class_id(self) -> str:
        """
        Return the Class+ID as a string.

        :return: Class+ID format
        """
        return f'{self.__class__.__name__}<"{self._id}">'

    def get_id(self) -> str:
        """
        Return the object ID.

        :return: Object ID
        """
        return self._id



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

SELECTION
Widget selection effect.
"""

__all__ = ['Selection']

import copy
import pygame
import pygame_menu

from pygame_menu.utils import assert_color

from pygame_menu._types import NumberType, ColorType, ColorInputType, Tuple2IntType, \
    Tuple4IntType, NumberInstance, Optional, Union


class Selection(object):
    """
    Widget selection effect class.

    .. note::

        All selection classes must be copyable.

    :param margin_left: Left margin
    :param margin_right: Right margin
    :param margin_top: Top margin
    :param margin_bottom: Bottom margin
    """
    color: ColorType
    color_bg: Optional[ColorType]
    margin_bottom: NumberType
    margin_left: NumberType
    margin_right: NumberType
    margin_top: NumberType
    widget_apply_font_color: bool

    def __init__(
        self,
        margin_left: NumberType,
        margin_right: NumberType,
        margin_top: NumberType,
        margin_bottom: NumberType
    ) -> None:
        assert isinstance(margin_left, NumberInstance)
        assert isinstance(margin_right, NumberInstance)
        assert isinstance(margin_top, NumberInstance)
        assert isinstance(margin_bottom, NumberInstance)
        assert margin_left >= 0, 'left margin of widget selection cannot be negative'
        assert margin_right >= 0, 'right margin of widget selection cannot be negative'
        assert margin_top >= 0, 'top margin of widget selection cannot be negative'
        assert margin_bottom >= 0, 'bottom margin of widget selection cannot be negative'

        self.color = (0, 0, 0)  # Main color of the selection effect
        self.color_bg = None
        self.margin_bottom = margin_bottom
        self.margin_left = margin_left
        self.margin_right = margin_right
        self.margin_top = margin_top
        self.widget_apply_font_color = True  # Widgets apply "selected_color" if selected

    def margin_xy(self, x: NumberType, y: NumberType) -> 'Selection':
        """
        Set margins at left-right / top-bottom.

        :param x: Left-Right margin in px
        :param y: Top-Bottom margin in px
        :return: Self reference
        """
        assert isinstance(x, NumberInstance) and x >= 0
        assert isinstance(y, NumberInstance) and y >= 0
        self.margin_left = x
        self.margin_right = x
        self.margin_top = y
        self.margin_bottom = y
        return self

    def zero_margin(self) -> 'Selection':
        """
        Makes selection margin zero.

        :return: Self reference
        """
        self.margin_top = 0
        self.margin_left = 0
        self.margin_right = 0
        self.margin_bottom = 0
        return self

    def copy(self) -> 'Selection':
        """
        Creates a deep copy of the object.

        :return: Copied selection effect
        """
        return copy.deepcopy(self)

    def __copy__(self) -> 'Selection':
        """
        Copy method.

        :return: Copied selection
        """
        return self.copy()

    def set_color(self, color: ColorInputType) -> 'Selection':
        """
        Set the selection effect color.

        :param color: Selection color
        :return: Self reference
        """
        self.color = assert_color(color)
        return self

    def set_background_color(self, color: Union[ColorInputType, 'pygame_menu.BaseImage']) -> 'Selection':
        """
        Set the selection background color. It will replace the background color of the widget
        if selected.

        :param color: Background color
        :return: Self reference
        """
        self.color_bg = color
        if not isinstance(color, pygame_menu.BaseImage):
            self.color_bg = assert_color(self.color_bg)
        return self

    def get_background_color(self) -> Optional[Union[ColorType, 'pygame_menu.BaseImage']]:
        """
        Return the background color.

        :return: Background color or None
        """
        return self.color_bg

    def get_margin(self) -> Tuple4IntType:
        """
        Return the top, left, bottom and right margins of the selection.

        :return: Tuple of (top, left, bottom, right) margins in px
        """
        return int(self.margin_top), int(self.margin_left), int(self.margin_bottom), int(self.margin_right)

    def get_xy_margin(self) -> Tuple2IntType:
        """
        Return the x/y margins of the selection.

        :return: Margin tuple on x-axis and y-axis (x, y) in px
        """
        return int(self.margin_left + self.margin_right), int(self.margin_top + self.margin_bottom)

    def get_width(self) -> int:
        """
        Return the selection width as sum of left and right margins.

        :return: Width in px
        """
        _, l, _, r = self.get_margin()
        return l + r

    def get_height(self) -> int:
        """
        Return the selection height as sum of top and bottom margins.

        :return: Height in px
        """
        t, _, b, _ = self.get_margin()
        return t + b

    def inflate(
        self,
        rect: 'pygame.Rect', inflate: Optional[Tuple2IntType] = None
    ) -> 'pygame.Rect':
        """
        Grow or shrink the rectangle size according to margins.

        :param rect: Rect object
        :param inflate: Extra border inflate
        :return: Inflated rect
        """
        if inflate is None:
            inflate = (0, 0)
        assert isinstance(rect, pygame.Rect)
        return pygame.Rect(
            int(rect.x - self.margin_left - inflate[0] / 2),
            int(rect.y - self.margin_top - inflate[1] / 2),
            int(rect.width + self.margin_left + self.margin_right + inflate[0]),
            int(rect.height + self.margin_top + self.margin_bottom + inflate[1])
        )

    def draw(self, surface: 'pygame.Surface', widget: 'pygame_menu.widgets.Widget') -> 'Selection':
        """
        Draw the selection.

        :param surface: Surface to draw
        :param widget: Widget object
        :return: Self reference
        """
        raise NotImplementedError('override is mandatory')



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

FONTS
Menu fonts.
"""

__all__ = [

    # Fonts path included in resources
    'FONT_8BIT',
    'FONT_BEBAS',
    'FONT_COMIC_NEUE',
    'FONT_DIGITAL',
    'FONT_FRANCHISE',
    'FONT_FIRACODE',
    'FONT_FIRACODE_BOLD',
    'FONT_FIRACODE_BOLD_ITALIC',
    'FONT_FIRACODE_ITALIC',
    'FONT_HELVETICA',
    'FONT_MUNRO',
    'FONT_NEVIS',
    'FONT_OPEN_SANS',
    'FONT_OPEN_SANS_BOLD',
    'FONT_OPEN_SANS_ITALIC',
    'FONT_OPEN_SANS_LIGHT',
    'FONT_PT_SERIF',
    'FONT_EXAMPLES',

    # Typing
    'FontType',
    'FontInstance',

    # Utils
    'assert_font',
    'get_font'

]

from pathlib import Path
from typing import Union, Optional, Any, Dict, Tuple
import os.path as path

import pygame.font as __font

# Available fonts path
__fonts_path__ = path.join(path.dirname(path.abspath(__file__)), 'resources', 'fonts', '{0}')

FONT_8BIT = __fonts_path__.format('8bit.ttf')
FONT_BEBAS = __fonts_path__.format('bebas.ttf')
FONT_COMIC_NEUE = __fonts_path__.format('comic_neue.ttf')
FONT_DIGITAL = __fonts_path__.format('digital.ttf')
FONT_FIRACODE = __fonts_path__.format('FiraCode-Regular.ttf')
FONT_FIRACODE_BOLD = __fonts_path__.format('FiraCode-Bold.ttf')
FONT_FIRACODE_BOLD_ITALIC = __fonts_path__.format('FiraMono-BoldItalic.ttf')
FONT_FIRACODE_ITALIC = __fonts_path__.format('FiraMono-Italic.ttf')
FONT_FRANCHISE = __fonts_path__.format('franchise.ttf')
FONT_HELVETICA = __fonts_path__.format('helvetica.ttf')
FONT_MUNRO = __fonts_path__.format('munro.ttf')
FONT_NEVIS = __fonts_path__.format('nevis.ttf')
FONT_OPEN_SANS = __fonts_path__.format('opensans_regular.ttf')
FONT_OPEN_SANS_BOLD = __fonts_path__.format('opensans_bold.ttf')
FONT_OPEN_SANS_ITALIC = __fonts_path__.format('opensans_italic.ttf')
FONT_OPEN_SANS_LIGHT = __fonts_path__.format('opensans_light.ttf')
FONT_PT_SERIF = __fonts_path__.format('ptserif_regular.ttf')

FONT_EXAMPLES = (FONT_8BIT, FONT_BEBAS, FONT_COMIC_NEUE, FONT_DIGITAL, FONT_FRANCHISE,
                 FONT_HELVETICA, FONT_MUNRO, FONT_NEVIS, FONT_OPEN_SANS,
                 FONT_OPEN_SANS_BOLD, FONT_OPEN_SANS_ITALIC, FONT_OPEN_SANS_LIGHT,
                 FONT_PT_SERIF, FONT_FIRACODE, FONT_FIRACODE_BOLD, FONT_FIRACODE_ITALIC,
                 FONT_FIRACODE_BOLD_ITALIC)

FontType = Union[str, __font.Font, Path]
FontInstance = (str, __font.Font, Path)

# Stores font cache
_cache: Dict[Tuple[FontType, int], '__font.Font'] = {}


def assert_font(font: Any) -> None:
    """
    Asserts if the given object is a font type.

    :param font: Font object
    """
    assert isinstance(font, FontInstance), \
        'value must be a font type (str, Path, pygame.Font)'


def get_font(name: FontType, size: int) -> '__font.Font':
    """
    Return a :py:class:`pygame.font.Font` object from a name or file.

    :param name: Font name or path
    :param size: Font size in px
    :return: Font object
    """
    assert_font(name)
    assert isinstance(size, int)

    font: Optional['__font.Font']
    if isinstance(name, __font.Font):
        font = name
        return font

    else:
        name = str(name)

        if name == '':
            raise ValueError('font name cannot be empty')

        if size <= 0:
            raise ValueError('font size cannot be lower or equal than zero')

        # Font is not a file, then use a system font
        if not path.isfile(name):
            font_name = name
            name = __font.match_font(font_name)

            if name is None:  # Show system available fonts
                from difflib import SequenceMatcher
                from random import randrange
                system_fonts = __font.get_fonts()

                # Get the most similar example
                most_similar = 0
                most_similar_index = 0
                for i in range(len(system_fonts)):
                    # noinspection PyArgumentEqualDefault
                    sim = SequenceMatcher(None, system_fonts[i], font_name).ratio()
                    if sim > most_similar:
                        most_similar = sim
                        most_similar_index = i
                sys_font_sim = system_fonts[most_similar_index]
                sys_suggestion = f'system font "{font_name}" unknown, use "{sys_font_sim}" instead'
                sys_message = 'check system fonts with pygame.font.get_fonts() function'

                # Get examples
                examples_number = 3
                examples = []
                j = 0
                for i in range(len(system_fonts)):
                    font_random = system_fonts[randrange(0, len(system_fonts))]
                    if font_random not in examples:
                        examples.append(font_random)
                        j += 1
                    if j >= examples_number:
                        break
                examples.sort()
                fonts_random = ', '.join(examples)
                sys_message_2 = f'some examples: {fonts_random}'

                # Raise the exception
                raise ValueError(f'{sys_suggestion}\n{sys_message}\n{sys_message_2}')

        # Try to load the font
        font = None
        if (name, size) in _cache:
            return _cache[(name, size)]
        try:
            font = __font.Font(name, size)
        except IOError:
            pass

        # If font was not loaded throw an exception
        if font is None:
            raise IOError(f'font file "{font}" cannot be loaded')
        _cache[(name, size)] = font
        return font
